# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2007 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime
from string import Template
from thread import get_ident, allocate_lock
from time import strptime

# Import from itools
from itools.datatypes import String
from itools.http import Response
from itools.i18n import AcceptLanguageType
from itools.uri import get_reference



class FormError(StandardError):
    """Raised when a form is invalid (missing or invalid fields).
    """

    def __init__(self, missing=None, invalid=None):
          self.missing = missing or []
          self.invalid = invalid or []


    def __str__(self):
        return '%s field(s) are missing and %s field(s) are invalid' \
               % (len(self.missing), len(self.invalid))



class Context(object):

    user = None
    object = None


    def __init__(self, request):
        self.request = request
        self.response = Response()


    def init(self):
        """To process a request it must be loaded, in the first place.
        """
        request = self.request
        if request.has_header('X-Forwarded-Host'):
            host = request.get_header('X-Forwarded-Host')
        elif request.has_header('Host'):
            host = request.get_header('Host')
        else:
            # XXX We should return a 400 response with HTTP 1.1
            # XXX What to do with 1.0?
            host = None

        if request.has_header('X_FORWARDED_PROTO'):
            scheme = request.get_header('X_FORWARDED_PROTO')
        else:
            # By default http
            scheme = 'http'

        # The requested uri
        reference = '%s://%s%s' % (scheme, host, request.request_uri)
        self.uri = get_reference(reference)

        # Split the path into path and method ("a/b/c/;view")
        path = request.request_uri.path
        if path and path[-1].name == '':
            self.path = path[:-1]
            self.method = path[-1].params[0]
        else:
            self.path = path
            self.method = request.method

        # Language negotiation
        headers = request.headers
        if 'accept-language' in headers:
            # FIXME Done this way the programmer may modify the request object
            # TODO The 'Accept-Language' header should be deserialized here,
            # not in the 'Request' object.
            self.accept_language = headers['accept-language']
        else:
            self.accept_language = AcceptLanguageType.decode('')


    #######################################################################
    # API / Redirect
    #######################################################################
    def redirect(self, reference, status=302):
        reference = self.uri.resolve(reference)
        self.response.redirect(reference, status)


    def come_back(self, message, goto=None, keep=[], **kw):
        """This is a handy method that builds a URI object from some
        parameters.  It exists to make short some common patterns.
        """
        # By default we come back to the referrer
        if goto is None:
            goto = self.request.referrer
            # Replace goto if no referrer
            if goto is None:
                uri = str(self.uri)
                if '/;' in uri:
                    goto = uri.split('/;')[0]
                    goto = get_reference(goto)
        elif isinstance(goto, str):
            goto = get_reference(goto)
        # Preserve some form values
        form = {}
        for key, value in self.request.form.items():
            # Be robust
            if not key:
                continue
            # Omit methods
            if key[0] == ';':
                continue
            # Omit files
            if isinstance(value, tuple) and len(value) == 3:
                continue
            # Keep form field
            if (keep is True) or (key in keep):
                form[key] = value
        if form:
            goto = goto.replace(**form)
        # Translate the source message
        if message:
            message = self.object.gettext(message)
            message = Template(message).substitute(kw)
            return goto.replace(message=message)
        return goto


    #######################################################################
    # API / Forms
    #######################################################################
    def get_form_keys(self):
        return self.request.form.keys()


    # TODO For the next major release, change the method signature, it
    # should be:
    #
    #   get_form_value(name, type=None, default=None)
    #
    def get_form_value(self, name, default=None, type=String):
        request = self.request

        # Figure out the default value
        is_multiple = getattr(type, 'multiple', False)
        if default is None:
            default = type.default
            if is_multiple and not isinstance(default, list):
                default = []

        # Missing
        is_mandatory = getattr(type, 'mandatory', False)
        is_missing = not request.has_parameter(name)
        if is_missing:
            # Mandatory: raise an error
            if is_mandatory and is_missing:
                raise FormError(missing=[name])
            # Optional: return the default value
            return default

        # Multiple values
        if is_multiple:
            value = request.get_parameter(name)
            if not isinstance(value, list):
                value = [value]
            try:
                values = [ type.decode(x) for x in value ]
            except:
                raise FormError(invalid=[name])
            # Check the values are valid
            for value in values:
                if not type.is_valid(value):
                    raise FormError(invalid=[name])
            return value

        # Single value
        value = request.get_parameter(name)
        if isinstance(value, list):
            value = value[0]
        try:
            value = type.decode(value)
        except:
            raise FormError(invalid=[name])

        # We consider that if the type deserializes the value to None, then
        # we must use the default.
        if value is None:
            return default

        # We consider a blank string to be a missing value (FIXME not
        # reliable).
        is_blank = isinstance(value, (str, unicode)) and not value.strip()
        if is_mandatory and is_blank:
            raise FormError(missing=[name])

        if not type.is_valid(value):
            raise FormError(invalid=[name])
        return value


    # FIXME Obsolete since 0.20.4, to be removed by the next major release
    def get_form_values(self, name, default=[], type=None):
        request = self.request
        if request.has_parameter(name):
            value = request.get_parameter(name)
            if not isinstance(value, list):
                value = [value]

            if type is None:
                return value
            return [ type.decode(x) for x in value ]

        return default


    def has_form_value(self, name):
        return self.request.has_parameter(name)


    def build_form_namespace(self, schema):
        """This utility method builds a namespace suitable to use to produce
        an HTML form. Its input data is a dictionnary that defines the form
        variables to consider:

          {'toto': Unicode(mandatory=True, multiple=False, default=u'toto'),
           'tata': Unicode(mandatory=True, multiple=False, default=u'tata')}

        Every element specifies the datatype of the field.
        The output is like:

            {<field name>: {'value': <field value>, 'class': <CSS class>}
             ...}
        """
        # Figure out whether the form has been submit or not (FIXME This
        # heuristic is not reliable)
        submit = set(self.get_form_keys()) & set(schema.keys())

        # Build the namespace
        namespace = {}
        for name in schema:
            datatype = schema[name]
            cls = []
            if getattr(datatype, 'mandatory', False):
                cls.append('field_required')
            if submit:
                try:
                    value = self.get_form_value(name, type=datatype)
                except FormError:
                    value = self.get_form_value(name)
                    cls.append('missing')
            else:
                value = datatype.default
            cls = ' '.join(cls) or None
            namespace[name] = {'name': name, 'value': value, 'class': cls}
        return namespace


    def check_form_input(self, schema):
        """Form checks the request form and collect inputs consider the
        schema.  This method also checks the request form and raise an
        FormError if there is something wrong (a mandatory field is missing,
        or a value is not valid) or None if everything is ok.

        Its input data is a list (fields) that defines the form variables to
          {'toto': Unicode(mandatory=True, multiple=False, default=u'toto'),
           'tata': Unicode(mandatory=True, multiple=False, default=u'tata')}
        """
        values = {}
        invalid = []
        missing = []
        for name in schema:
            datatype = schema[name]
            try:
                value = self.get_form_value(name, type=datatype)
            except FormError, error:
                value = self.get_form_value(name)
                missing.extend(error.missing)
                invalid.extend(error.invalid)
            values[name] = value
        if missing or invalid:
            raise FormError(missing, invalid)
        return values


    #######################################################################
    # API / Cookies
    #######################################################################
    def get_cookie(self, name, type=None):
        request, response = self.request, self.response
        # Get the value
        cookie = response.get_cookie(name)
        if cookie is None:
            cookie = request.get_cookie(name)
            if cookie is None:
                value = None
            else:
                value = cookie.value
        else:
            # Check expiration time
            value = cookie.value
            expires = cookie.expires
            if expires is not None:
                expires = expires[5:-4]
                expires = strptime(expires, '%d-%b-%y %H:%M:%S')
                year, month, day, hour, min, sec, kk, kk, kk = expires
                expires = datetime(year, month, day, hour, min, sec)
                if expires < datetime.now():
                    value = None

        if type is None:
            return value

        # Deserialize
        if value is None:
            return type.default
        return type.decode(value)


    def has_cookie(self, name):
        return self.get_cookie(name) is not None


    def set_cookie(self, name, value, **kw):
        self.response.set_cookie(name, value, **kw)


    def del_cookie(self, name):
        self.response.del_cookie(name)



###########################################################################
# One context per thread
###########################################################################
contexts = {}
contexts_lock = allocate_lock()


def set_context(context):
    ident = get_ident()

    contexts_lock.acquire()
    try:
        contexts[ident] = context
    finally:
        contexts_lock.release()


def has_context():
    return get_ident() in contexts


def get_context():
    return contexts.get(get_ident())


def del_context():
    ident = get_ident()

    contexts_lock.acquire()
    try:
        del contexts[ident]
    finally:
        contexts_lock.release()
